/*
 * This is the High Performance Python Profiler portion of HotShot.
 *
 * This is an enhanced version of source file .../Modules/_hotshot.c
 * from Python release 2.5.2 using the high-resolution timer where
 * available and otherwise falling back to gettimeofday as before.
 *
 * All profile times are recorded in 64-bit ticks.  The tick length
 * can be obtained from new info item 'timer-frequency' logged as
 * ticks per second.  Two other files .../Lib/hotshot/log.py and
 * .../Lib/hotshot/stats.py have been updated and use the new info
 * item 'timer-frequency'.
 *
 * The high-resolution timer is used only if the C compiler does
 * support 64-bit long, either as PY_LONG_LONG or if SIZEOF_LONG
 * equals 8.  Several changes were made to log (delta) ticks as
 * 64-bit ints when needed.
 *
 * In addition, this file includes a number of other fixes for
 * potential problems with the original one.  For example, string
 * lengths can not exceed the buffer size, infinite recursion is
 * avoided, options are set before writing the header info into
 * the log file, int is coerced to long where needed, malloc is
 * only called when needed and usage of type Py_ssize_t/size_t.
 *
 * Tested with Python 2.5.2 on 32-bit MacOS X (both Intel and PPC),
 * 32-bit CentOS, 64-bit RHEL 3u7 and 64-bit Solaris 10.
 *
 * A high-resolution timer may be useful elsewhere.  Perhaps as
 * another function in the time module and/or in a core file like
 * .../Python/ceval.c.
 */

#include "Python.h"
#include "code.h"
#include "eval.h"
#include "frameobject.h"
#include "structmember.h"

/*
 * Which timer to use should be made more configurable,
 * that should not be difficult.  This will do for now.
 */

#ifndef inline
# if defined(__GNUC__) && !defined(__cplusplus)
#   define inline __inline
# else
#   define inline
# endif
#endif

#ifdef  _GETTICKS
# error _GETTICKS already defined
#endif
#ifdef  _GETRUSAGE
# error _GETRUSAGE already defined
#endif

#ifdef MS_WINDOWS
# include <windows.h>

# ifdef HAVE_DIRECT_H
#   include <direct.h>    /* for getcwd() */
# endif

    typedef __int64 _ticks_t;  /* XXX unsigned would be better */

    static inline _ticks_t
    _getticks (void)
    {
        LARGE_INTEGER t; \
        QueryPerformanceCounter(&t); \
        return ((_ticks_t)t.QuadPart);
    }

#   define _GETTICKS   0  /* no _getticks, _gettimeofday */
#   define _GETRUSAGE  0  /* no _getrusage */

#else
  /* use hires timer only if a 64-bit long type is
     available, otherwise use _gettimeofday as timer */
# ifdef HAVE_LONG_LONG
    typedef unsigned PY_LONG_LONG _ticks_t;
#   define _GETTICKS               SIZEOF_LONG_LONG
# else
    typedef unsigned long _ticks_t;
#   define _GETTICKS       SIZEOF_LONG
# endif

  /* XXX HAVE_GETHRTIME should probably be defined in
     pyconfig.h.  Solaris and HP-UX provide a function
     gethrtime returning a 64-bit time stamp with nano-
     second accuracy (and using a very low overhead
     system call on Solaris UltraSPARC). */
# if _GETTICKS == 8 && defined(HAVE_GETHRTIME)  /* Solaris and HP-UX only */
#   include <time.h>
#   define  _getticks  (_ticks_t) gethrtime

# elif _GETTICKS == 8 && defined(__GNUC__) && (defined(__i386__) || defined(__x86_64__))
    static inline _ticks_t
    _getticks (void)
    {
       _ticks_t t;
#     ifdef rdtscll
         /* rdtscll() is defined in <asm/msr.h>
            of certain, recent Linux distro's */
        rdtscll(t);
#     else
        /* XXX tested */
        unsigned int hi, lo;
        asm volatile ("rdtsc" : "=a" (lo), "=d" (hi));
        t = (((_ticks_t)hi) << 32) | (_ticks_t)lo;
#     endif
        return (t);
    }

# elif _GETTICKS == 8 && defined (__GNUC__) && (defined(__powerpc64__) || defined(__POWERPC64__))
#   warning This module is untested on 64-bit PPC.
    static /* inline */ _ticks_t
    _getticks (void)
    {
       _ticks_t t;
        asm volatile ("mftb %0" : "=r" (t));
        return (t);
    }

# elif _GETTICKS == 8 && defined (__GNUC__) && (defined(__ppc__) || defined(__POWERPC__))
    static /* inline */ _ticks_t
    _getticks (void)
    {
       _ticks_t t;
        unsigned int hi, hi2, lo;
        do {  /* see also .../Python/ceval.c */
            asm volatile ("mftbu %0 \n"
                          "mftb  %1 \n"
                          "mftbu %2 \n" : "=r"(hi), "=r"(lo), "=r"(hi2));
        } while (__builtin_expect(hi != hi2, 0));
#     if 0  /* XXX tested, see also .../Python/ceval.c */
        ((unsigned int*)(t))[0] = hi;
        ((unsigned int*)(t))[1] = lo;
#     else  /* XXX tested */
        t = (((_ticks_t)hi) << 32) | (_ticks_t)lo;
#     endif
        return (t);
    }

# else  /* no hires timer or 64-bit long */
#   undef  _GETTICKS
#   define _GETTICKS   0
    /* use _gettimeofday returning usecs */
#   define _getticks  _gettimeofday
# endif

# if defined(PYOS_OS2) || defined(__QNX__) || defined(__VMS)
#   include <sys/time.h>
#   define _GETRUSAGE  0

# else
#   include <sys/resource.h>
#   include <sys/times.h>
#   define _GETRUSAGE  1

    static _ticks_t
    _getrusage (void)
    {
        struct rusage ru;
        getrusage(RUSAGE_SELF, &ru);
        return ((_ticks_t)(ru.ru_utime.tv_usec + ru.ru_stime.tv_usec) + (_ticks_t)1000000 *
                (_ticks_t)(ru.ru_utime.tv_sec  + ru.ru_stime.tv_sec));
    }
# endif

# ifndef HAVE_GETTIMEOFDAY
#   error "Need gettimeofday() function in _hotshot.c"
# endif

    static inline _ticks_t
    _gettimeofday (void)  /* in usec ticks */
    {
        struct timeval tv;
#     ifdef GETTIMEOFDAY_NO_TZ
        gettimeofday(&tv)
#     else
        gettimeofday(&tv, (struct timezone *)NULL);
#     endif
        return ((_ticks_t)tv.tv_usec + (_ticks_t)1000000 * (_ticks_t)tv.tv_sec);
    }

#endif

#define BUFFERSIZE 10240

#if defined(PYOS_OS2) && defined(PYCC_GCC)
#define PATH_MAX 260
#endif

#if defined(__sgi) && _COMPILER_VERSION>700 && !defined(PATH_MAX)
/* fix PATH_MAX not being defined with MIPSPro 7.x
   if mode is ANSI C (default) */
#define PATH_MAX 1024
#endif

#ifndef PATH_MAX
#   ifdef MAX_PATH
#       define PATH_MAX MAX_PATH
#   elif defined (_POSIX_PATH_MAX)
#       define PATH_MAX _POSIX_PATH_MAX
#   else
#       error "Need a defn. for PATH_MAX in _hotshot.c"
#   endif
#endif

#if 0
  typedef unsigned char _byte_t;
#else  /* const prefix */
# define _byte_t  unsigned char
#endif

typedef struct {
    PyObject_HEAD
    PyObject *filemap;
    PyObject *logfilename;
    Py_ssize_t index;
    FILE *logfp;
    int active;
    int err;  /* saved error */
    int lineevents;
    int linetimings;
    int frametimings;
    int next_fileno;
   _ticks_t prev_ticks;
   _byte_t buffer[BUFFERSIZE];
} ProfilerObject;

typedef struct {
    PyObject_HEAD
    PyObject *info;
    FILE *logfp;
    int linetimings;
    int frametimings;
} LogReaderObject;


static PyObject *ProfilerError = NULL;

static  double  timer_frequency  = 0.0;
static  long    timer_resolution = 0L;
static _ticks_t timer_tolerance  = (_ticks_t)0;  /* error tolerance */
static  long    timeofday_diff   = 0L;
static  long    rusage_diff      = 0L;  /* -1L if _getrusage n/a */
#if _GETTICKS
    /* start ticks and clock time */
    static _ticks_t ticks0       = (_ticks_t)0;
    static _ticks_t timeofday0   = (_ticks_t)0;
#endif


/* Close profiler output or log reader input file.
 */
static inline void
do_close(FILE **logfp_ptr)
{
    FILE *logfp = *logfp_ptr;
    if (logfp != NULL) {
        fclose(logfp);
       *logfp_ptr = NULL;
    }
}


/* Set a profiler or log reader error.
 */

#define ERR_NONE        0
#define ERR_RETURN     -1  /* for tracer_callback */
#define ERR_ACTIVE     -2
#define ERR_BUFFER     -3
#define ERR_CLOSED     -4
#define ERR_EOF        -5
#define ERR_EXCEPTION  -6
#define ERR_INACTIVE   -7
#define ERR_INDEX      -8
#define ERR_IO         -9
#define ERR_ITERATE    -10
#define ERR_NEGATIVE   -11
#define ERR_READER     -12
#define ERR_RECORD     -13
#define ERR_TICKS      -14
#define ERR_WHATMOD    -15
#define ERR_WRITING    -16

static int
set_error(int err, LogReaderObject *self)
{
    switch (err) {
    case ERR_NONE:      return 0;

    case ERR_ACTIVE:    PyErr_SetString(ProfilerError, "profiler already active");
                        break;

    case ERR_BUFFER:    PyErr_SetString(PyExc_AssertionError, "length exceeds buffer size");
                        break;

    case ERR_CLOSED:    PyErr_SetString(PyExc_ValueError, "profiler already closed");
                        break;

    case ERR_EOF:       if (self)
                            do_close(&self->logfp);
                        PyErr_SetString(PyExc_EOFError, "end of file with incomplete profile record");
                        break;

    case ERR_EXCEPTION: break;  /* error already set */

    case ERR_INACTIVE:  PyErr_SetString(ProfilerError, "profiler not active");
                        break;

    case ERR_INDEX:     PyErr_SetString(PyExc_IndexError, "no more events in log");
                        break;

    case ERR_IO:        break;  /* error already set */

    case ERR_ITERATE:   PyErr_SetString(ProfilerError, "cannot iterate over closed log reader");
                        break;

    case ERR_NEGATIVE:  PyErr_SetString(PyExc_AssertionError, "length value negative");
                        break;

    case ERR_READER:    PyErr_SetString(PyExc_ValueError, "log reader already closed");
                        break;

    case ERR_RECORD:    PyErr_SetString(PyExc_ValueError, "invalid profile record type");
                        break;

    case ERR_TICKS:     PyErr_SetString(ProfilerError, "timer exceeds backward tolerance");
                        break;

    case ERR_WHATMOD:   PyErr_SetString(PyExc_AssertionError, "modifier exceeds size");
                        break;

    case ERR_WRITING:   PyErr_SetString(ProfilerError, "cannot write entire log");
                        break;

    default:            PyErr_SetString(PyExc_RuntimeError, "unexpected profile error");
                        break;
    }
    return ERR_RETURN;  /* for tracer_callback */
}


/* Log File Format
 * ---------------
 *
 * The log file consists of a sequence of variable-length records.
 * Each record is identified with a record type identifier in two
 * bits of the first byte.  The two bits are the "least significant"
 * bits of the byte.
 *
 * Low bits:    Opcode:        Meaning:
 *       0x00         ENTER     enter a frame
 *       0x01          EXIT     exit a frame
 *       0x02        LINENO     execution moved onto a different line
 *       0x03         OTHER     more bits are needed to deecode
 *
 * If the type is OTHER, the record is not packed so tightly, and the
 * remaining bits are used to disambiguate the record type.  These
 * records are not used as frequently so compaction is not an issue.
 * Each of the first three record types has a highly tailored
 * structure that allows it to be packed tightly.
 *
 * The OTHER records have the following identifiers:
 *
 * First byte:  Opcode:        Meaning:
 *       0x13      ADD_INFO     define a key/value pair
 *       0x23   DEFINE_FILE     define an int->filename mapping
 *       0x33    LINE_TIMES     indicates if LINENO events have tdeltas
 *       0x43   DEFINE_FUNC     define a (fileno,lineno)->funcname mapping
 *       0x53   FRAME_TIMES     indicates if ENTER/EXIT events have tdeltas
 *
 * Packed integers and ticks
 *
 * "Packed integers" (PI) are non-negative integer values encoded as
 * a sequence of bytes.  Each byte is encoded such that the most
 * significant bit is set if the next byte is also part of the
 * integer.  Each byte provides bits to the least-significant end of
 * the result; the accumulated value must be shifted up to place the
 * new bits into the result.
 *
 * "Modified packed integers" (MPI) are packed integers where only a
 * portion of the first byte is used.  In the rest of the specification,
 * these are referred to as "MPI(n,name)", where "n" is the number of
 * bits discarded from the least-signicant positions of the byte, and
 * "name" is a name being given to those "discarded" bits, since they
 * are a field themselves.
 *
 * "Packed ticks" (PT) and "modified packed ticks" (MPT) are like
 * "packed" resp. "modified packed integers" but can contain a non-
 * negative integer value of up to 64 bits.
 *
 * ENTER records:
 *
 *      MPI(2,type)  fileno          -- type is 0x00
 *      PI           lineno
 *      PT           tdelta          -- iff frame times are enabled
 *
 * EXIT records
 *
 *      MTI(2,type)  tdelta          -- type is 0x01; tdelta will be 0
 *                                      if frame times are disabled
 *
 * LINENO records
 *
 *      MPI(2,type)  lineno          -- type is 0x02
 *      PT           tdelta          -- iff LINENO includes it
 *
 * ADD_INFO records
 *
 *      BYTE         type            -- always 0x13
 *      PI           len1            -- length of first string
 *      BYTE         string1[len1]   -- len1 bytes of string data
 *      PI           len2            -- length of second string
 *      BYTE         string2[len2]   -- len2 bytes of string data
 *
 * DEFINE_FILE records
 *
 *      BYTE         type            -- always 0x23
 *      PI           fileno
 *      PI           len             -- length of filename
 *      BYTE         filename[len]   -- len bytes of string data
 *
 * DEFINE_FUNC records
 *
 *      BYTE         type            -- always 0x43
 *      PI           fileno
 *      PI           lineno
 *      PI           len             -- length of funcname
 *      BYTE         funcname[len]   -- len bytes of string data
 *
 * LINE_TIMES records
 *
 * This record can be used only before the start of ENTER/EXIT/LINENO
 * records.  If have_tdelta is true, LINENO records will include the
 * tdelta field, otherwise it will be omitted.  If this record is not
 * given, LINENO records will not contain the tdelta field.
 *
 *      BYTE         type            -- always 0x33
 *      BYTE         have_tdelta     -- 0 if LINENO does *not* have
 *                                      timing information
 * FRAME_TIMES records
 *
 * This record can be used only before the start of ENTER/EXIT/LINENO
 * records.  If have_tdelta is true, ENTER and EXIT records will
 * include the tdelta field, otherwise it will be omitted.  If this
 * record is not given, ENTER and EXIT records will contain the tdelta
 * field.
 *
 *      BYTE         type            -- always 0x53
 *      BYTE         have_tdelta     -- 0 if ENTER/EXIT do *not* have
 *                                      timing information
 */

#define WHAT_ENTER        0x00
#define WHAT_EXIT         0x01
#define WHAT_LINENO       0x02
#define WHAT_OTHER        0x03  /* only used in decoding */
#define WHAT_ADD_INFO     0x13
#define WHAT_DEFINE_FILE  0x23
#define WHAT_LINE_TIMES   0x33
#define WHAT_DEFINE_FUNC  0x43
#define WHAT_FRAME_TIMES  0x53

#define NIBBITS      7  /* bits per nibble */
#define NIBCONT     (1 << NIBBITS)  /* continue bit */
#define NIBMASK     (NIBCONT - 1)   /* nibble mask */

#define PISIZE      ((sizeof(int)      + NIBBITS - 1) / NIBBITS)  /* in nibbles */
#define PTSIZE      ((sizeof(_ticks_t) + NIBBITS - 1) / NIBBITS)  /* in nibbles */
#define MPISIZE     (PISIZE + 1)  /* modified packed integer */
#define MPTSIZE     (PTSIZE + 1)  /* modified packed ticks */
#define STRSIZE(n)  (n + PISIZE)  /* string of n chars */

#define MODSIZE      2  /* in bits */
#define MODBITS     (NIBBITS - MODSIZE)
#define MODMASK     ((1 << MODBITS) - 1)


/* The log reader... */

static int
unpack_nibbles(LogReaderObject *self, _byte_t* buf, size_t n)
{
    int c;
    int i = 0;
    do {
        c = fgetc(self->logfp);
        if (c == EOF)
            return ERR_EOF;
        else if (n <= (size_t)i)
            return ERR_BUFFER;
        buf[i++] = (_byte_t)(c & NIBMASK);
    } while (c & NIBCONT);
    return i;
}

/* Unpack a packed integer; if "discard" is non-zero, unpack
 * a modified packed integer with "discard" discarded bits.
 */
static int
unpack_packed_int(LogReaderObject *self, int *pvalue, int modified)
{
   _byte_t buf[MPISIZE+2];
    int i = unpack_nibbles(self, buf, sizeof(buf));
    if (i > 0) {
        int v = (int)buf[--i];
        while (i-- > 0) {
            v <<= NIBBITS;
            v  |= (int)buf[i];
        }
        *pvalue = modified ? (v >> MODSIZE) : v;
        return ERR_NONE;
    }
    return i;
}

static int
unpack_packed_ticks(LogReaderObject *self, _ticks_t *pvalue, int modified)
{
   _byte_t buf[MPTSIZE+2];
    int i = unpack_nibbles(self, buf, sizeof(buf));
    if (i > 0) {
       _ticks_t v = (_ticks_t)buf[--i];
        while (i-- > 0) {
            v <<= NIBBITS;
            v  |= (_ticks_t)buf[i];
        }
        *pvalue = modified ? (v >> MODSIZE) : v;
        return ERR_NONE;
    }
    return i;
}

static int
unpack_bool(LogReaderObject *self, int *pvalue)
{
    int c = fgetc(self->logfp);
    if (c == EOF)
        return ERR_EOF;
    *pvalue = c ? 1 : 0;
    return ERR_NONE;
}

/* Unpack a string, which is encoded as a packed integer giving the
 * length of the string, followed by the string data.
 */
static int
unpack_string(LogReaderObject *self, PyObject **pvalue)
{
    int i;
    char str[256];
    char *buf = str;

    int len = 0;
    int err = unpack_packed_int(self, &len, 0);
    if (err)
        return err;

    if (len < 0)
        return ERR_NEGATIVE;

    if (sizeof(str) < (size_t)len) {
        buf = (char *)malloc((size_t)len);
        if (!buf) {
            PyErr_NoMemory();
            return ERR_EXCEPTION;
        }
    }

    for (i = 0; i < len; i++) {
        int ch = fgetc(self->logfp);
        buf[i] = ch;
        if (ch == EOF) {
            err = ERR_EOF;
            goto finally;
        }
    }
    *pvalue = PyString_FromStringAndSize(buf, (Py_ssize_t)len);
    if (*pvalue == NULL)
        err = ERR_EXCEPTION;

 finally:
    if (buf != str) {
        free(buf);
    }
    return err;
}

/* Unpack key-value pair and append value
 * to the list of info dict item [key]..
 */
static int
unpack_add_info(LogReaderObject *self)
{
    PyObject *key   = NULL;
    PyObject *value = NULL;

    int err = unpack_string(self, &key);
    if (!err) {
        err = unpack_string(self, &value);
        if (!err) {
            PyObject *list = PyDict_GetItem(self->info, key);
            if (list == NULL) {
                list = PyList_New(0);
                if (list == NULL) {
                    err = ERR_EXCEPTION;
                    goto finally;
                }
                if (PyDict_SetItem(self->info, key, list)) {
                    Py_DECREF(list);
                    err = ERR_EXCEPTION;
                    goto finally;
                }
                Py_DECREF(list);
            }
            if (PyList_Append(list, value))
                err = ERR_EXCEPTION;
        }
    }
 finally:
    Py_XDECREF(key);
    Py_XDECREF(value);
    return err;
}

static PyObject *
logreader_tp_iternext(LogReaderObject *self)
{
    int c, what, err;
    int lineno = -1;
    int fileno = -1;
   _ticks_t tdelta = ~ (_ticks_t)0;  /* -1 */
    PyObject *s1 = NULL;
    PyObject *s2 = NULL;
    PyObject *result = NULL;

    if (self->logfp == NULL) {
        set_error(ERR_ITERATE, self);
        return NULL;
    }

restart:
    /* decode the record type */
    if ((c = fgetc(self->logfp)) == EOF) {
        do_close(&self->logfp);
        return NULL;
    }
    what = c & WHAT_OTHER;
    if (what == WHAT_OTHER)
        what = c; /* need all the bits for type */
    else
        ungetc(c, self->logfp); /* type byte includes packed int */

    switch (what) {
    case WHAT_ENTER:
        err = unpack_packed_int(self, &fileno, MODSIZE);
        if (!err) {
             err = unpack_packed_int(self, &lineno, 0);
             if (!err && self->frametimings)
                  err = unpack_packed_ticks(self, &tdelta, 0);
        }
        break;
    case WHAT_EXIT:
        err = unpack_packed_ticks(self, &tdelta, MODSIZE);
        break;
    case WHAT_LINENO:
        err = unpack_packed_int(self, &lineno, MODSIZE);
        if (!err && self->linetimings)
             err = unpack_packed_ticks(self, &tdelta, 0);
        break;
    case WHAT_ADD_INFO:
        err = unpack_add_info(self);
        break;
    case WHAT_DEFINE_FILE:
        err = unpack_packed_int(self, &fileno, 0);
        if (!err) {
             err = unpack_string(self, &s1);
             if (!err) {
                 Py_INCREF(Py_None);
                 s2 = Py_None;
             }
        }
        break;
    case WHAT_DEFINE_FUNC:
        err = unpack_packed_int(self, &fileno, 0);
        if (!err) {
             err = unpack_packed_int(self, &lineno, 0);
             if (!err)
                  err = unpack_string(self, &s1);
        }
        break;
    case WHAT_LINE_TIMES:
        err = unpack_bool(self, &self->linetimings);
        if (!err)
            goto restart;
        break;
    case WHAT_FRAME_TIMES:
        err = unpack_bool(self, &self->frametimings);
        if (!err)
            goto restart;
        break;
    default:
        err = ERR_RECORD;
    }
    if (err)
        set_error(err, self);
    else {
        result = PyTuple_New(4);
        if (result != NULL) {
            PyTuple_SET_ITEM(result, 0, PyInt_FromLong((long)what));
            PyTuple_SET_ITEM(result, 2, PyInt_FromLong((long)fileno));
            if (s1 == NULL)
#ifdef HAVE_LONG_LONG
# ifdef MS_WINDOWS
                s1 = PyLong_FromLongLong((PY_LONG_LONG)tdelta);
# else
                s1 = PyLong_FromUnsignedLongLong(tdelta);
# endif
#else
# ifdef MS_WINDOWS
                s1 = PyLong_FromLong((tdelta < (_ticks_t)LONG_MAX) ? (long)tdelta : LONG_MAX);
# else
                s1 = PyLong_FromUnsignedLong(tdelta);
# endif
#endif
            PyTuple_SET_ITEM(result, 1, s1);
            if (s2 == NULL)
                s2 = PyInt_FromLong((long)lineno);
            PyTuple_SET_ITEM(result, 3, s2);
        }
    }
#if 0
    if (1) {
       _byte_t b0 = self->buffer[self->index];
       _byte_t b1 = self->buffer[self->index + 1];
        if (b0 & 1) {
            /* This is a line-number event. */
            what = PyTrace_LINE;
            lineno = ((b0 & ~1) << 7) + b1;
            self->index += 2;
        }
        else {
            what = (b0 & 0x0E) >> 1;
            tdelta = ((b0 & 0xF0) << 4) + b1;
            if (what == PyTrace_CALL) {
                /* we know there's a 2-byte file ID & 2-byte line number */
                fileno = ((self->buffer[self->index + 2] << 8)
                         + self->buffer[self->index + 3]);
                lineno = ((self->buffer[self->index + 4] << 8)
                         + self->buffer[self->index + 5]);
                self->index += 6;
            }
            else
                self->index += 2;
        }
    }
#endif
    return result;
}

static PyObject *
logreader_sq_item(LogReaderObject *self, Py_ssize_t index)
{
    PyObject *result = logreader_tp_iternext(self);
    if (result == NULL && !PyErr_Occurred())
        set_error(ERR_INDEX, self);
    return result;
}

PyDoc_STRVAR(logreader_close__doc__,
"close()\n"
"Close the log file and prevent reading of additional records.");

static PyObject *
logreader_close(LogReaderObject *self, PyObject *args)
{
    do_close(&self->logfp);
    Py_INCREF(Py_None);

    return Py_None;
}

static void
logreader_dealloc(LogReaderObject *self)
{
    do_close(&self->logfp);
    Py_XDECREF(self->info);
    PyObject_Del(self);
}

PyDoc_STRVAR(logreader_fileno__doc__,
"fileno() -> file descriptor\n"
"Returns the file descriptor for the log file, if open.\n"
"Raises a ValueError if the log file is closed.");

static PyObject *
logreader_fileno(LogReaderObject *self)
{
    if (self->logfp == NULL) {
        set_error(ERR_READER, self);
        return NULL;
    }
    return PyInt_FromLong((long)fileno(self->logfp));
}


/* The profiler... */

static int do_stop(ProfilerObject *self, int raiserr);  /* forward */

/* Dump data to the log file.
 */
static int
flush_data(ProfilerObject *self, Py_ssize_t res, int raiserr)
{
    int i;
    Py_ssize_t n = self->index;

    if (self->logfp == NULL)
        return ERR_CLOSED;

    /* try to write the data several times */
    for (i = 4; n > 0 && i > 0; i--) {
        Py_ssize_t w = (Py_ssize_t)fwrite(self->buffer, 1, (size_t)n, self->logfp);
        if (w < 0)
            return ERR_NEGATIVE;

        self->index = n = n - w;
        if (n > 0 && w > 0)  /* partial write */
            memcpy(self->buffer, &self->buffer[w], (size_t)n);

        if (w == 0 || fflush(self->logfp)) {  /* write or flush error */
            if (raiserr) {
                char *s = PyString_AsString(self->logfilename);
                PyErr_SetFromErrnoWithFilename(PyExc_IOError, s);
            }
            if (self->active)
                do_stop(self, 0);  /* ignore errors */
            return ERR_IO;
        }
    }

    if (n > res) {  /* residual exceeded */
        if (self->active)
            do_stop(self, raiserr);
        return ERR_WRITING;
    }
    return ERR_NONE;
}

/* Check available buffer space.
 */
static inline int
make_space(ProfilerObject *self, Py_ssize_t len)
{
    return (((self->index + len) < BUFFERSIZE) ? ERR_NONE
           :               (len  > BUFFERSIZE) ? ERR_BUFFER
           : flush_data(self, BUFFERSIZE - len, 1));
}

static inline void
copy_byte(ProfilerObject *self, _byte_t value)
{
    self->buffer[self->index] = value;
    self->index++;
}

static inline void
copy_bytes(ProfilerObject *self, const _byte_t *values, Py_ssize_t len)
{
    if (len > 0) {
        memcpy(self->buffer + self->index, values, (size_t)len);
        self->index += len;
    }
}

/* Encode an integer.
 */
static inline int
pack_packed_int(ProfilerObject *self, int value)
{
    int v = value;  /* XXX unsigned? */
    do {
       _byte_t p = (_byte_t)(v & NIBMASK);
        v >>= NIBBITS;
        if (v)
            p |= NIBCONT;
        copy_byte(self, p);
    } while (v);

    return ERR_NONE;
}

static inline int
pack_packed_ticks(ProfilerObject *self, _ticks_t value)
{
   _ticks_t v = value;
    do {
       _byte_t p = (_byte_t)(v & NIBMASK);
        v >>= NIBBITS;
        if (v)
            p |= NIBCONT;
        copy_byte(self, p);
    } while (v);

    return ERR_NONE;
}

/* Encode a modified packed integer, with a subfield
 * of MODSIZE bits containing the value of 'what'.
 */
static inline int
pack_modified_packed_int(ProfilerObject *self, int value, _byte_t what)
{
    if (what < (1 << MODSIZE)) {
       _byte_t p = (_byte_t)(value & MODMASK);
       _byte_t b = (value != (int)p) ? NIBCONT : 0;

        copy_byte(self, what | b | (p << MODSIZE));
        return (b ? pack_packed_int(self, value >> MODBITS) : ERR_NONE);
    }
    return ERR_WHATMOD;
}

static inline int
pack_modified_packed_ticks(ProfilerObject *self, _ticks_t value, _byte_t what)
{
    if (what < (1 << MODSIZE)) {
       _byte_t p = (_byte_t)(value & MODMASK);
       _byte_t b = (value != (_ticks_t)p) ? NIBCONT : 0;

        copy_byte(self, what | b | (p << MODSIZE));
        return (b ? pack_packed_ticks(self, value >> MODBITS) : ERR_NONE);
    }
    return ERR_WHATMOD;
}

/* Encode a packed delta ticks value, modified if non-zero.
 */
static int
pack_delta_ticks(ProfilerObject *self, _ticks_t ticks, int modified, _byte_t what)
{
    int d = 0;
   _ticks_t p = self->prev_ticks;
    /* time can go backwards on some MP systems
       or by NTP or hires timer may wrap around */
    if (ticks > p) {
        self->prev_ticks = ticks;
        ticks -= p;
        if (ticks >= (_ticks_t)INT_MAX)
            return (modified ? pack_modified_packed_ticks(self, ticks, what)
                             :          pack_packed_ticks(self, ticks));
        d = (int)ticks;
    }
    else if ((p - ticks) > timer_tolerance)
        /* correct timings */
        self->prev_ticks = ticks;
    else  /* too far backward */
        return ERR_TICKS;
    return (modified ? pack_modified_packed_int(self, d, what)
                     :          pack_packed_int(self, d));
}

static int
pack_string(ProfilerObject *self, const char *str, Py_ssize_t len)
{
    int err = ERR_BUFFER;
    if (len >= 0) {
        err = make_space(self, STRSIZE(len));
        if (!err) {
             err = pack_packed_int(self, (int)len);
             if (!err)
                 copy_bytes(self, (const _byte_t *)str, len);
        }
    }
    return err;
}

static int
pack_bool(ProfilerObject *self, int value, _byte_t what)
{
    int err = make_space(self, 2);
    if (!err) {
        copy_byte(self, what);
        copy_byte(self, value ? 1 : 0);
    }
    else if (!self->err)
        self->err = err;
    return err;
}

static int
pack_add_info(ProfilerObject *self, const char *str1, const char *str2)
{
    int err = ERR_BUFFER;
    Py_ssize_t len1 = (Py_ssize_t)strlen(str1);
    Py_ssize_t len2 = (Py_ssize_t)strlen(str2);

    if (0 <= len1 && len1 < BUFFERSIZE &&
        0 <= len2 && len2 < BUFFERSIZE) {
        err = make_space(self, 1 + STRSIZE(len1) + STRSIZE(len2));
        if (!err) {
            copy_byte(self, WHAT_ADD_INFO);
            err = pack_string(self, str1, len1);
            if (!err)
                 err = pack_string(self, str2, len2);
        }
    }
    if (err && !self->err)
        self->err = err;
    return err;
}

static int
pack_define_file(ProfilerObject *self, int fileno,
                 const char *filename)
{
    int err = ERR_BUFFER;
    Py_ssize_t len = (Py_ssize_t)strlen(filename);

    if (0 <= len && len < BUFFERSIZE) {
        err = make_space(self, 1 + PISIZE + STRSIZE(len));
        if (!err) {
            copy_byte(self, WHAT_DEFINE_FILE);
            err = pack_packed_int(self, fileno);
            if (!err)
                 err = pack_string(self, filename, len);
        }
    }
    return err;
}

static int
pack_define_func(ProfilerObject *self, int fileno, int lineno,
                 const char *funcname)
{
    int err = ERR_BUFFER;
    Py_ssize_t len = (Py_ssize_t)strlen(funcname);

    if (0 <= len && len < BUFFERSIZE) {
        err = make_space(self, 1 + PISIZE*2 + STRSIZE(len));
        if (!err) {
            copy_byte(self, WHAT_DEFINE_FUNC);
            err = pack_packed_int(self, fileno);
            if (!err) {
                 err = pack_packed_int(self, lineno);
                 if (!err)
                      err = pack_string(self, funcname, len);
            }
        }
    }
    return err;
}

static inline int
pack_enter(ProfilerObject *self, int fileno, int lineno, _ticks_t ticks)
{
    int err = make_space(self, MPISIZE + PISIZE + PTSIZE);
    if (!err) {
         err = pack_modified_packed_int(self, fileno, WHAT_ENTER);
         if (!err) {
              err = pack_packed_int(self, lineno);
              if (!err && self->frametimings)
                   err = pack_delta_ticks(self, ticks, 0, 0);
         }
    }
    return err;
}

static inline int
pack_exit(ProfilerObject *self, _ticks_t ticks)
{
    int err = make_space(self, MPTSIZE);
    if (!err) {
        if (self->frametimings)
            err = pack_delta_ticks(self, ticks, MODSIZE, WHAT_EXIT);
        else
            copy_byte(self, WHAT_EXIT);
    }
    return err;
}

static inline int
pack_lineno(ProfilerObject *self, int lineno, _ticks_t ticks)
{
    int err = make_space(self, MPISIZE + PTSIZE);
    if (!err) {
         err = pack_modified_packed_int(self, lineno, WHAT_LINENO);
         if (!err && self->linetimings)
              err = pack_delta_ticks(self, ticks, 0, 0);
    }
    return err;
}

static inline int
get_fileno(ProfilerObject *self, PyCodeObject *fcode)
{
    /* This is only used for ENTER events. */

    PyObject *obj;
    PyObject *dict;
    int fileno, err;

    obj = PyDict_GetItem(self->filemap, fcode->co_filename);
    if (obj == NULL) {
        /* first sighting of this file */
        dict = PyDict_New();
        if (dict == NULL) {
            return ERR_EXCEPTION;
        }
        fileno = self->next_fileno;
        obj = Py_BuildValue("iN", fileno, dict);
        if (obj == NULL) {
            return ERR_EXCEPTION;
        }
        if (PyDict_SetItem(self->filemap, fcode->co_filename, obj)) {
            Py_DECREF(obj);
            return ERR_EXCEPTION;
        }
        self->next_fileno++;
        Py_DECREF(obj);
        err = pack_define_file(self, fileno,
                               PyString_AS_STRING(fcode->co_filename));
        if (err)
            return err;
    }
    else {
        /* already know this ID */
        fileno = PyInt_AS_LONG(PyTuple_GET_ITEM(obj, 0));
        dict = PyTuple_GET_ITEM(obj, 1);
        err = ERR_NONE;
    }
    /* make sure we save a function name for this (fileno, lineno) */
    obj = PyInt_FromLong((long)fcode->co_firstlineno);
    if (obj == NULL) {
        /* We just won't have it saved; too bad. */
        PyErr_Clear();
    }
    else {
        PyObject *name = PyDict_GetItem(dict, obj);
        if (name == NULL) {
            err = pack_define_func(self, fileno, fcode->co_firstlineno,
                                   PyString_AS_STRING(fcode->co_name));
            if (!err && PyDict_SetItem(dict, obj, fcode->co_name))
                 err = ERR_EXCEPTION;
        }
        Py_DECREF(obj);
    }
    return (err ? err : fileno);
}


/* The workhorse:  the profiler callback function. */

static int
tracer_callback(ProfilerObject *self, PyFrameObject *frame, int what,
                PyObject *arg)
{
    int fileno, err;

    switch (what) {
    case PyTrace_CALL:
        fileno = get_fileno(self, frame->f_code);
        if (fileno < 0)
            return set_error(fileno, NULL);
        err = pack_enter(self, fileno, frame->f_code->co_firstlineno,
                         self->frametimings ? _getticks() : (_ticks_t)0);
        break;

    case PyTrace_RETURN:
        err = pack_exit(self, self->frametimings ? _getticks() : (_ticks_t)0);
        break;

    case PyTrace_LINE:  /* we only get these events if we asked for them */
        err = pack_lineno(self, frame->f_lineno,
                          self->linetimings ? _getticks() : (_ticks_t)0);
        break;

    default:
        /* ignore PyTrace_EXCEPTION */
        return 0;
    }
    return (err ? set_error(err, NULL) : 0);
}


/* A couple of useful helper functions. */

/* Calibrate several times since sometimes the
 * first doesn't give the lowest values, and
 * we're really trying to determine the lowest.
 */
static void
calibrate (int cycles)
{
    int i;
    long d;
   _ticks_t t1, t2;
#ifdef MS_WINDOWS
    LARGE_INTEGER f;
    QueryPerformanceFrequency(&f);
    timer_frequency = (double)f.QuadPart;
    if (!timer_resolution) {
        t1 = _getticks();
        do {
            t2 = _getticks();
        } while (t2 <= t1);
        timer_resolution = (long)(t2 - t1);  /* ticks */
    }
#else
    for (i = cycles; i-- > 0; ) {
        t1 = _gettimeofday();
        do {
            t2 = _gettimeofday();
        } while (t2 <= t1);
        d = (long)(t2 - t1);
        if (timeofday_diff > d || !timeofday_diff) {
            timeofday_diff = d;
        }
    }
# if _GETRUSAGE
    for (i = cycles; i-- > 0; ) {
        t1 = _getrusage();
        do {
            t2 = _getrusage();
        } while (t2 <= t1);
        d = (long)(t2 - t1);
        if (rusage_diff > d || !rusage_diff) {
            rusage_diff = d;
        }
    }
# else
    rusage_diff = -1L;  /* n/a */
# endif
# if _GETTICKS
    if (!timer_resolution) {
        t1 = _getticks();
        do {
            t2 = _getticks();
        } while (t2 <= t1);
        timer_resolution = (long)(t2 - t1);  /* ticks */
    }
    /* determine hires timer frequency by waiting at least
       0.25 secs since the very start and divide the number
       of ticks since by the elapsed clock time in seconds */
    t1 = timeofday0 + (_ticks_t)250000;
    do {
        t2 = _gettimeofday();
    } while (t2 <= t1);
    t1 = _getticks();
    timer_frequency = 1.0e6 * (double)(t1 - ticks0) / (double)(t2 - timeofday0);
# else
    timer_resolution = timeofday_diff;
    /* clock ticks may not be in usecs but
       _gettimeofday return values are */
    timer_frequency  = 1000000.0;  /* usecs */
# endif
    /* error tolerance */
    timer_tolerance = (_ticks_t)timer_resolution << 2;
#endif
}

static int
do_start(ProfilerObject *self)
{
    int err;

    if (self->active)
        err = ERR_ACTIVE;
    else if (self->logfp == NULL)
        err = ERR_CLOSED;
    else {
        self->active = 1;
        self->prev_ticks = _getticks();
        if (self->lineevents)
            PyEval_SetTrace((Py_tracefunc)tracer_callback, (PyObject *)self);
        else
            PyEval_SetProfile((Py_tracefunc)tracer_callback, (PyObject *)self);
        return 1;  /* OK */
    }

    set_error(err, NULL);
    return 0;  /* NAK */
}

static int
do_stop(ProfilerObject *self, int raiserr)
{
    if (self->active) {
        self->active = 0;
        if (self->lineevents)
            PyEval_SetTrace(NULL, NULL);
        else
            PyEval_SetProfile(NULL, NULL);
    }
    if (self->index > 0) {
        /* Best effort to dump out any remaining data. */
        int err = flush_data(self, 0, raiserr);
        if (err) {
            if (raiserr)
                set_error(err, NULL);
            return 0;  /* NAK */
        }
    }
    return 1;  /* OK */
}


/* Profiler object interface methods. */

PyDoc_STRVAR(addinfo__doc__,
"addinfo(key, value)\n"
"Insert an ADD_INFO record into the log.");

static PyObject *
profiler_addinfo(ProfilerObject *self, PyObject *args)
{
    char *key, *value;
    PyObject *result = NULL;

    if (self->logfp == NULL)
        set_error(ERR_CLOSED, NULL);
    else if (PyArg_ParseTuple(args, "ss:addinfo", &key, &value)) {
        int err = pack_add_info(self, key, value);
        if (err)
            set_error(err, NULL);
        else {
            result = Py_None;
            Py_INCREF(result);
        }
    }
    return result;
}

PyDoc_STRVAR(close__doc__,
"close()\n"
"Shut down this profiler and close the log file even if active.");

static PyObject *
profiler_close(ProfilerObject *self)
{
    PyObject *result = NULL;

    if (do_stop(self, 1)) {
        result = Py_None;
        Py_INCREF(result);
    }
    do_close(&self->logfp);
    return result;
}

#define fileno__doc__ logreader_fileno__doc__

static PyObject *
profiler_fileno(ProfilerObject *self)
{
    if (self->logfp == NULL) {
        set_error(ERR_CLOSED, NULL);
        return NULL;
    }
    return PyInt_FromLong((long)fileno(self->logfp));
}

PyDoc_STRVAR(runcall__doc__,
"runcall(callable[, args[, kw]]) -> callable()\n"
"Profile a specific function call and return the result of that call.");

static PyObject *
profiler_runcall(ProfilerObject *self, PyObject *args)
{
    PyObject *callable;
    PyObject *callargs = NULL;
    PyObject *callkwds = NULL;
    PyObject *result = NULL;

    if (PyArg_UnpackTuple(args, "runcall", 1, 3,
                         &callable, &callargs, &callkwds)) {
        if (do_start(self)) {
            result = PyEval_CallObjectWithKeywords(callable, callargs, callkwds);
            /* XXX keep existing error iff any error is set */
            do_stop(self, PyErr_Occurred() ? 0 : 1);
        }
    }
    return result;
}

PyDoc_STRVAR(runcode__doc__,
"runcode(code, globals[, locals])\n"
"Execute and profile a code object and return the result.\n"
"If locals is omitted, globals is used for locals as well.");

static PyObject *
profiler_runcode(ProfilerObject *self, PyObject *args)
{
    PyCodeObject *code;
    PyObject *globals;
    PyObject *locals = NULL;
    PyObject *result = NULL;

    if (PyArg_ParseTuple(args, "O!O!|O:runcode",
                         &PyCode_Type, &code,
                         &PyDict_Type, &globals,
                         &locals)) {
        if (locals == NULL || locals == Py_None)
            locals = globals;
        if (!PyDict_Check(locals))
            PyErr_SetString(PyExc_TypeError,
                           "locals must be a dictionary or None");
        else if (do_start(self)) {
            result = PyEval_EvalCode(code, globals, locals);
            /* XXX keep existing error iff any error is set */
            do_stop(self, PyErr_Occurred() ? 0 : 1);
        }
    }
    return result;
}

PyDoc_STRVAR(start__doc__,
"start()\n"
"Install this profiler for the current thread.");

static PyObject *
profiler_start(ProfilerObject *self, PyObject *args)
{
    PyObject *result = NULL;

    if (do_start(self)) {
        result = Py_None;
        Py_INCREF(result);
    }
    return result;
}

PyDoc_STRVAR(stop__doc__,
"stop()\n"
"Remove this profiler from the current thread.");

static PyObject *
profiler_stop(ProfilerObject *self, PyObject *args)
{
    PyObject *result = NULL;

    if (!self->active)
        set_error(ERR_INACTIVE, NULL);
    else if (do_stop(self, 1)) {
        result = Py_None;
        Py_INCREF(result);
    }
    return result;
}


/* Python API support. */

static void
profiler_dealloc(ProfilerObject *self)
{
    do_stop(self, 0);  /* ignore errors */
    do_close(&self->logfp);
    Py_XDECREF(self->filemap);
    Py_XDECREF(self->logfilename);
    PyObject_Del((PyObject *)self);
}

static PyMethodDef profiler_methods[] = {
    {"addinfo", (PyCFunction)profiler_addinfo, METH_VARARGS, addinfo__doc__},
    {"close",   (PyCFunction)profiler_close,   METH_NOARGS,  close__doc__},
    {"fileno",  (PyCFunction)profiler_fileno,  METH_NOARGS,  fileno__doc__},
    {"runcall", (PyCFunction)profiler_runcall, METH_VARARGS, runcall__doc__},
    {"runcode", (PyCFunction)profiler_runcode, METH_VARARGS, runcode__doc__},
    {"start",   (PyCFunction)profiler_start,   METH_NOARGS,  start__doc__},
    {"stop",    (PyCFunction)profiler_stop,    METH_NOARGS,  stop__doc__},
    {NULL, NULL}
};

static PyMemberDef profiler_members[] = {
    {"frametimings", T_INT, offsetof(ProfilerObject, frametimings), READONLY},
    {"lineevents",   T_INT, offsetof(ProfilerObject, lineevents),   READONLY},
    {"linetimings",  T_INT, offsetof(ProfilerObject, linetimings),  READONLY},
    {NULL}
};

static PyObject *
profiler_get_closed(ProfilerObject *self, void *closure)
{
    PyObject *result = (self->logfp == NULL) ? Py_True : Py_False;
    Py_INCREF(result);
    return result;
}

static PyGetSetDef profiler_getsets[] = {
    {"closed", (getter)profiler_get_closed, NULL,
     PyDoc_STR("True if the profiler has already been closed.")},
    {NULL}
};


PyDoc_STRVAR(profiler_object__doc__,
"High-performance profiler object.\n"
"\n"
"Methods:\n"
"\n"
"close():      Stop profiling and close the log file.\n"
"fileno():     Return the file descriptor of the log file.\n"
"runcall():    Call a single function with profiling enabled.\n"
"runcode():    Execute a code object with profiling enabled.\n"
"start():      Install this profiler for the current thread.\n"
"stop():       Remove this profiler from the current thread.\n"
"\n"
"Attributes (read-only):\n"
"\n"
"closed:       True if the profiler has already been closed.\n"
"frametimings: True if enter/exit events contain timing information.\n"
"lineevents:   True if line events are reported to the profiler.\n"
"linetimings:  True if line events contain timing information.");

static PyTypeObject ProfilerType = {
    PyObject_HEAD_INIT(NULL)
    0,					/* ob_size		*/
    "_hotshot.ProfilerType",		/* tp_name		*/
    (int)sizeof(ProfilerObject),	/* tp_basicsize		*/
    0,					/* tp_itemsize		*/
    (destructor)profiler_dealloc,	/* tp_dealloc		*/
    0,					/* tp_print		*/
    0,					/* tp_getattr		*/
    0,					/* tp_setattr		*/
    0,					/* tp_compare		*/
    0,					/* tp_repr		*/
    0,					/* tp_as_number		*/
    0,					/* tp_as_sequence	*/
    0,					/* tp_as_mapping	*/
    0,					/* tp_hash		*/
    0,					/* tp_call		*/
    0,					/* tp_str		*/
    PyObject_GenericGetAttr,		/* tp_getattro		*/
    0,					/* tp_setattro		*/
    0,					/* tp_as_buffer		*/
    Py_TPFLAGS_DEFAULT,			/* tp_flags		*/
    profiler_object__doc__,		/* tp_doc		*/
    0,					/* tp_traverse		*/
    0,					/* tp_clear		*/
    0,					/* tp_richcompare	*/
    0,					/* tp_weaklistoffset	*/
    0,					/* tp_iter		*/
    0,					/* tp_iternext		*/
    profiler_methods,			/* tp_methods		*/
    profiler_members,			/* tp_members		*/
    profiler_getsets,			/* tp_getset		*/
    0,					/* tp_base		*/
    0,					/* tp_dict		*/
    0,					/* tp_descr_get		*/
    0,					/* tp_descr_set		*/
};


static PyMethodDef logreader_methods[] = {
    {"close",  (PyCFunction)logreader_close,  METH_NOARGS, logreader_close__doc__},
    {"fileno", (PyCFunction)logreader_fileno, METH_NOARGS, logreader_fileno__doc__},
    {NULL, NULL}
};

static PyMemberDef logreader_members[] = {
    {"info",         T_OBJECT, offsetof(LogReaderObject, info), RO,
     PyDoc_STR("Dictionary mapping informational keys to lists of values.")},
    {"frametimings", T_INT, offsetof(LogReaderObject, frametimings), READONLY},
    {"linetimings",  T_INT, offsetof(LogReaderObject, linetimings),  READONLY},
    {NULL}
};


PyDoc_STRVAR(logreader__doc__,
"logreader(filename) --> log-iterator\n"
"Create a log reader for the given profiler log file.\n"
"\n"
"Methods:\n"
"\n"
"close():      Close the log file.\n"
"fileno():     Return the file descriptor of the log file.\n"
"\n"
"Attributes (read-only):\n"
"\n"
"closed:       True if the log file has already been closed.\n"
"info:         Dictionary mapping informational keys to lists of values.\n"
"frametimings: True if the log file contains enter/exit events with timing information.\n"
"linetimings:  True if the log file contains line events with timing information.");


static PySequenceMethods logreader_as_sequence = {
    0,					/* sq_length */
    0,					/* sq_concat */
    0,					/* sq_repeat */
    (ssizeargfunc)logreader_sq_item,	/* sq_item */
    0,					/* sq_slice */
    0,					/* sq_ass_item */
    0,					/* sq_ass_slice */
    0,					/* sq_contains */
    0,					/* sq_inplace_concat */
    0,					/* sq_inplace_repeat */
};

static PyObject *
logreader_get_closed(LogReaderObject *self, void *closure)
{
    PyObject *result = (self->logfp == NULL) ? Py_True : Py_False;
    Py_INCREF(result);
    return result;
}

static PyGetSetDef logreader_getsets[] = {
    {"closed", (getter)logreader_get_closed, NULL,
     PyDoc_STR("True if the log file has already been closed.")},
    {NULL}
};

static PyTypeObject LogReaderType = {
    PyObject_HEAD_INIT(NULL)
    0,					/* ob_size		*/
    "_hotshot.LogReaderType",		/* tp_name		*/
    (int)sizeof(LogReaderObject),	/* tp_basicsize		*/
    0,					/* tp_itemsize		*/
    (destructor)logreader_dealloc,	/* tp_dealloc		*/
    0,					/* tp_print		*/
    0,					/* tp_getattr		*/
    0,					/* tp_setattr		*/
    0,					/* tp_compare		*/
    0,					/* tp_repr		*/
    0,					/* tp_as_number		*/
    &logreader_as_sequence,		/* tp_as_sequence	*/
    0,					/* tp_as_mapping	*/
    0,					/* tp_hash		*/
    0,					/* tp_call		*/
    0,					/* tp_str		*/
    PyObject_GenericGetAttr,		/* tp_getattro		*/
    0,					/* tp_setattro		*/
    0,					/* tp_as_buffer		*/
    Py_TPFLAGS_DEFAULT,			/* tp_flags		*/
    logreader__doc__,			/* tp_doc		*/
    0,					/* tp_traverse		*/
    0,					/* tp_clear		*/
    0,					/* tp_richcompare	*/
    0,					/* tp_weaklistoffset	*/
    PyObject_SelfIter,			/* tp_iter		*/
    (iternextfunc)logreader_tp_iternext,/* tp_iternext		*/
    logreader_methods,			/* tp_methods		*/
    logreader_members,			/* tp_members		*/
    logreader_getsets,			/* tp_getset		*/
    0,					/* tp_base		*/
    0,					/* tp_dict		*/
    0,					/* tp_descr_get		*/
    0,					/* tp_descr_set		*/
};

static PyObject *
hotshot_logreader(PyObject *unused, PyObject *args)
{
    char *filename;

    if (PyArg_ParseTuple(args, "s:logreader", &filename)) {
        LogReaderObject *self = PyObject_New(LogReaderObject, &LogReaderType);
        if (self == NULL)
            return (PyObject *)NULL;
        self->frametimings = 1;
        self->linetimings = 0;
        self->info = NULL;

        self->logfp = fopen(filename, "rb");
        if (self->logfp == NULL) {
            PyErr_SetFromErrnoWithFilename(PyExc_IOError, filename);
            goto finally;
        }
        self->info = PyDict_New();
        if (self->info == NULL)
            goto finally;

        /* read initial info */
        for (;;) {
            int c = fgetc(self->logfp);
            if (c == EOF) {
                set_error(ERR_EOF, self);
                goto finally;
            }
            else if (c == WHAT_ADD_INFO) {
                int err = unpack_add_info(self);
                if (err) {
                    set_error(err, self);
                    goto finally;
                }
            }
            else {  /* done */
                ungetc(c, self->logfp);
                return (PyObject *)self;
            }
        }

    finally:
        do_close(&self->logfp);
        Py_XDECREF(self->info);
        Py_DECREF(self);
    }
    return (PyObject *)NULL;
}


/* Return a string that represents the version number
 * without the extra cruft added by revision control,
 * even if the right options were given to "cvs export"
 * to make it not include the extra cruft.
 */
static char *
version_string(char *str, size_t n)
{
    static char *rcsid = "$Revision: 51218 $";
    char *rev = rcsid;
    int i = 0;

    while (*rev && !isdigit(Py_CHARMASK(*rev))) {
        ++rev;
    }
    while (*rev && *rev != ' ' && --n) {
        str[i++] = *rev++;
    }
    str[i] = '\0';
    return str;
}

/* Create a new profiler and configure it using the given options.
 */
static PyObject *
profiler_object(char* objectname, PyObject *args, char *logfilename,
                int frametimings, int lineevents, int linetimings)
{
    ProfilerObject *self;
    Py_ssize_t n, i;
    char *buf;
    char str[PATH_MAX];

    PyObject *path = PySys_GetObject("path");
    if (path == NULL || !PyList_Check(path)) {
        PyErr_SetString(PyExc_RuntimeError, "sys.path must be a list");
        return (PyObject *)NULL;
    }

    self = PyObject_New(ProfilerObject, &ProfilerType);
    if (self == NULL)
        return (PyObject *)NULL;
    self->lineevents = lineevents ? 1 : 0;
    self->linetimings = (lineevents && linetimings) ? 1 : 0;
    self->frametimings = frametimings ? 1 : 0;
    self->index = 0;
    self->active = 0;
    self->err = ERR_NONE;
    self->next_fileno = 0;
    self->prev_ticks = (_ticks_t)0;
    self->logfp = NULL;
    self->logfilename = NULL;
    self->filemap = PyDict_New();
    if (self->filemap == NULL)
        goto finally;
    self->logfilename = PyTuple_GET_ITEM(args, 0);
    if (self->logfilename == NULL)
        goto finally;
    Py_INCREF(self->logfilename);
    self->logfp = fopen(logfilename, "wb");
    if (self->logfp == NULL) {
        PyErr_SetFromErrnoWithFilename(PyExc_IOError, logfilename);
        goto finally;
    }

    /* Write a RFC 822-style header with various useful bits
     * of information to make the output easier to manage.
     */
    pack_add_info(self, "hotshot", objectname);  /* XXX added */
    pack_add_info(self, "hotshot-version",
                         version_string(str, sizeof(str)));
    pack_add_info(self, "requested-frame-timings",
                        (self->frametimings ? "yes" : "no"));
    pack_add_info(self, "requested-line-events",
                        (self->lineevents ? "yes" : "no"));
    pack_add_info(self, "requested-line-timings",
                        (self->linetimings ? "yes" : "no"));
    pack_add_info(self, "platform", Py_GetPlatform());
    pack_add_info(self, "executable", Py_GetProgramFullPath());
    buf = (char *)Py_GetVersion();
    if (buf == NULL)
        PyErr_Clear();
    else
        pack_add_info(self, "executable-version", buf);

    if (!timeofday_diff)
        calibrate(3);
    PyOS_snprintf(str, sizeof(str), "%ld", timer_resolution);
    pack_add_info(self, "timer-resolution", str);  /* XXX added */
    PyOS_snprintf(str, sizeof(str), "%.3f", timer_frequency);
    pack_add_info(self, "timer-frequency", str);  /* XXX added */
    PyOS_snprintf(str, sizeof(str), "%d", _GETTICKS);
    pack_add_info(self, "timer-hiresizeof", str);  /* XXX added */

#ifdef MS_WINDOWS
    PyOS_snprintf(str, sizeof(str), "%.0f", timer_frequency);
    pack_add_info(self, "reported-performance-frequency", str);
#else
    PyOS_snprintf(str, sizeof(str), "%ld", timeofday_diff);
    pack_add_info(self, "observed-interval-gettimeofday", str);
    PyOS_snprintf(str, sizeof(str), "%ld", rusage_diff);
    pack_add_info(self, "observed-interval-getrusage", str);
#endif
    pack_add_info(self, "current-directory", getcwd(str, sizeof(str)));

    /* add all sys.path items */
    n = PyList_GET_SIZE(path);
    for (i = 0; i < n; ++i) {
        PyObject *item = PyList_GET_ITEM(path, i);
        buf = item ? PyString_AsString(item) : NULL;
        if (buf == NULL) {
            buf = "<non-string-path-entry>";
            PyErr_Clear();
        }
        pack_add_info(self, "sys-path-entry", buf);
    }

    /* add timing options as last header records */
    pack_bool(self, self->frametimings, WHAT_FRAME_TIMES);
    pack_bool(self, self->linetimings,  WHAT_LINE_TIMES);

    if (!self->err)
        return (PyObject *)self;

    set_error(self->err, NULL);

finally:  /* some error occurred */
    do_close(&self->logfp);
    Py_XDECREF(self->filemap);
    Py_XDECREF(self->logfilename);
    Py_DECREF(self);
    return (PyObject *)NULL;
}

PyDoc_STRVAR(profiler__doc__,
"profiler(logfilename[, lineevents[, linetimes]]) -> profiler\n\
Create a new profiler object.");

static PyObject *
hotshot_profiler(PyObject *unused, PyObject *args)
{
    char *logfilename;
    int lineevents = 0;
    int linetimings = 1;

    if (PyArg_ParseTuple(args, "s|ii:profiler", &logfilename,
                         &lineevents, &linetimings))
        return profiler_object("profiler", args, logfilename, 1, lineevents, linetimings);
    return (PyObject *)NULL;
}

PyDoc_STRVAR(coverage__doc__,
"coverage(logfilename) -> profiler\n\
Returns a profiler that does not collect timing information.\n\
This is useful for building a coverage analysis tool.");

static PyObject *
hotshot_coverage(PyObject *unused, PyObject *args)
{
    char *logfilename;

    if (PyArg_ParseTuple(args, "s:coverage", &logfilename))
        return profiler_object("coverage", args, logfilename, 0, 1, 0);
    return (PyObject *)NULL;
}

PyDoc_VAR(resolution__doc__) = 
#ifdef MS_WINDOWS
PyDoc_STR(
"resolution() -> (performance-counter-ticks, update-frequency)\n"
"Return the resolution of the timer provided by the QueryPerformanceCounter()\n"
"function.  The first value is the smallest observed change, and the second\n"
"is the result of QueryPerformanceFrequency()."
)
#elif _GETTICKS
PyDoc_STR(
"resolution() -> (timer-resolution, timer-frequency)\n"
"Return the resolution (in ticks) and frequency (in ticks per second)\n"
"of the high-resolution timer."
)
#else
PyDoc_STR(
"resolution() -> (gettimeofday-usecs, getrusage-usecs)\n"
"Return the resolution of the timers provided by the gettimeofday() and\n"
"getrusage() system calls, or -1 if the call is not supported."
)
#endif
;

static PyObject *
hotshot_resolution(PyObject *self, PyObject *unused)
{
#ifdef MS_WINDOWS
    calibrate(1);
    return Py_BuildValue("ld", timer_resolution, timer_frequency);
#elif _GETTICKS
    calibrate(1);
    return Py_BuildValue("ld", timer_resolution, timer_frequency);
#else
    if (!timeofday_diff)
        calibrate(3);
    return Py_BuildValue("ll", timeofday_diff, rusage_diff);
#endif
}


static PyMethodDef functions[] = {
    {"coverage",   hotshot_coverage,   METH_VARARGS, coverage__doc__},
    {"profiler",   hotshot_profiler,   METH_VARARGS, profiler__doc__},
    {"logreader",  hotshot_logreader,  METH_VARARGS, logreader__doc__},
    {"resolution", hotshot_resolution, METH_NOARGS,  resolution__doc__},
    {NULL, NULL}
};


void
init_hotshot(void)
{
    char str[64];
    PyObject *module;

    LogReaderType.ob_type = &PyType_Type;
    ProfilerType.ob_type = &PyType_Type;

    module = Py_InitModule("_hotshot", functions);
    if (module != NULL) {
#if _GETTICKS
        /* get initial clock and hires time values */
        timeofday0 = _gettimeofday();
        ticks0     = _getticks();
#endif
        PyModule_AddStringConstant(module, "__version__",
                                   version_string(str, sizeof(str)));
        Py_INCREF(&LogReaderType);
        PyModule_AddObject(module, "LogReaderType",
                           (PyObject *)&LogReaderType);
        Py_INCREF(&ProfilerType);
        PyModule_AddObject(module, "ProfilerType",
                           (PyObject *)&ProfilerType);

        if (ProfilerError == NULL)
            ProfilerError = PyErr_NewException("hotshot.ProfilerError",
                                               NULL, NULL);
        if (ProfilerError != NULL) {
            Py_INCREF(ProfilerError);
            PyModule_AddObject(module, "ProfilerError", ProfilerError);
        }
        PyModule_AddIntConstant(module, "WHAT_ENTER",       WHAT_ENTER);
        PyModule_AddIntConstant(module, "WHAT_EXIT",        WHAT_EXIT);
        PyModule_AddIntConstant(module, "WHAT_LINENO",      WHAT_LINENO);
        PyModule_AddIntConstant(module, "WHAT_OTHER",       WHAT_OTHER);
        PyModule_AddIntConstant(module, "WHAT_ADD_INFO",    WHAT_ADD_INFO);
        PyModule_AddIntConstant(module, "WHAT_DEFINE_FILE", WHAT_DEFINE_FILE);
        PyModule_AddIntConstant(module, "WHAT_DEFINE_FUNC", WHAT_DEFINE_FUNC);
        PyModule_AddIntConstant(module, "WHAT_LINE_TIMES",  WHAT_LINE_TIMES);
        PyModule_AddIntConstant(module, "WHAT_FRAME_TIMES", WHAT_FRAME_TIMES);
    }
}
